import operator
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Optional

from report_writer.core.content.base import BaseContent
from report_writer.core.content.samples.getters import ContextContent
from report_writer.core.content.samples.statics import StaticContent


if TYPE_CHECKING:
    UnaryOperator = Callable[[Any], Any]
    BinaryOperator = Callable[[Any, Any], Any]


class UnaryOperatorContent(BaseContent):
    """
    Контент описывающий получение данных с помощью унарной операции.

    :param content: Контент
    :param operation: Функция-оператор
    """
    def __init__(
            self,
            content: BaseContent,
            operation: 'UnaryOperator'
    ) -> None:
        self._left_content: BaseContent = content
        self._operation: 'UnaryOperator' = operation

    def render(self, context: Any) -> Any:
        return self._operation(
            self._left_content.render(context)
        )


class BinaryOperatorContent(BaseContent):
    """
    Контент описывающий получение данных с помощью бинарной операции

    :param left_content: Контент для левого операнда
    :param right_content: Контент для правого операнда
    :param operation: Функция-оператор
    """
    def __init__(
            self,
            left_content: BaseContent,
            right_content: BaseContent,
            operation: 'BinaryOperator'
    ) -> None:
        self._left_content: BaseContent = left_content
        self._right_content: BaseContent = right_content
        self._operation: 'BinaryOperator' = operation

    def render(self, context: Any) -> Any:
        return self._operation(
            self._left_content.render(context),
            self._right_content.render(context)
        )


class Content(BaseContent):
    """
    Контент-обёртка для поддержки операций.

    Использование:

    .. code-block:: python

        # Унарный оператор:
        content = -Content()
        content.render(2)  # -2
        # Бинарный оператор:
        content = Content() + 3
        content.render(2)  # 5
        # Бинарный оператор с другим контентом:
        content = Content() + StaticContent(3)
        content.render(2)  # 5

    :param content: Оборачиваемый контент
    """

    default_base_content: BaseContent = ContextContent()

    def __init__(self, content: Optional[BaseContent] = None) -> None:
        if content is None:
            content = self.default_base_content
        self._content: BaseContent = content

    def _unary_operator(self, operation: 'UnaryOperator') -> 'Content':
        """
        Создание контента с использованием унарного оператора

        :param operation: Функция-оператор
        """
        return Content(UnaryOperatorContent(self, operation))

    def _binary_operator(
            self,
            operation: 'BinaryOperator',
            right: Any
    ) -> 'Content':
        """
        Создание контента с использованием бинарного оператора

        :param operation: Функция-оператор
        :param right: Контент для правого операнда
        """
        if not isinstance(right, BaseContent):
            # Если пришли произвольные данные оборачиваем их в StaticContent
            right_: BaseContent = StaticContent(right)
        else:
            right_ = right

        return Content(BinaryOperatorContent(self, right_, operation))

    def __pos__(self) -> 'Content':
        return self._unary_operator(operator.__pos__)

    def __neg__(self) -> 'Content':
        return self._unary_operator(operator.__neg__)

    def __abs__(self) -> 'Content':
        return self._unary_operator(operator.__abs__)

    def __invert__(self) -> 'Content':
        # !!! Используется как замена not, т.к. последнего нет среди
        # методов объекта, а сама операция менее востребована чем :code:`not`
        return self._unary_operator(operator.__not__)

    def invert(self) -> 'Content':
        # Замена __invert__ для инверсии (побитового дополнения)
        # вместо :code:`~C()` следует использовать :code:`C().invert()`)
        return self._unary_operator(operator.__invert__)

    def __add__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__add__, other)

    def __sub__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__sub__, other)

    def __mul__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__mul__, other)

    def __floordiv__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__floordiv__, other)

    def __truediv__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__truediv__, other)

    def __mod__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__mod__, other)

    def __pow__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__pow__, other)

    def __lt__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__lt__, other)

    def __le__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__le__, other)

    def __gt__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__gt__, other)

    def __ge__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__ge__, other)

    # Проверка типов не пройдёт, т.к. метод определён в object
    def __eq__(self, other: Any) -> 'Content':  # type: ignore
        return self._binary_operator(operator.__eq__, other)

    # Проверка типов не пройдёт, т.к. метод определён в object
    def __ne__(self, other: Any) -> 'Content':  # type: ignore
        return self._binary_operator(operator.__ne__, other)

    def __or__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__or__, other)

    def __and__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__and__, other)

    def __xor__(self, other: Any) -> 'Content':
        return self._binary_operator(operator.__xor__, other)

    def __rshift__(self, other: Any) -> 'Content':
        # !!! Используется как передача аргумента в функцию,
        # т.к. сам rshift используется редко
        # :code:`(C() >> ''.join).render(['1', '2', '3'])` => :code:`'123'`
        # как замена rshift можно передать отрицательное значение в lshift
        # :code:`(C() << -2).render(4)` => :code:`4 >> 2`
        return self._binary_operator(lambda arg, fn: fn(arg), other)

    def __lshift__(self, other: Any) -> 'Content':
        # При передаче отрицательного числа будет использоваться
        # как замена rshift
        # (C() << -2).render(4) => 4 >> 2
        return self._binary_operator(
            lambda left, right: (
                operator.__lshift__(left, right)
                if right > 0 else
                operator.__rshift__(left, -right)
            ),
            other
        )

    def render(self, context: Any) -> Any:
        return self._content.render(context)


# Сокращения для удобства использования
U = UnaryOperatorContent
B = BinaryOperatorContent
C = Content
