from operator import itemgetter
from unittest import TestCase

from report_writer.core.content.samples.getters import AttrContent
from report_writer.core.content.samples.getters import ContextContent
from report_writer.core.content.samples.getters import IndexContent
from report_writer.core.content.samples.getters import KeyContent
from report_writer.core.content.samples.getters import TemplateContent
from report_writer.core.content.samples.iterators import (
    BorderedContentIterator
)
from report_writer.core.content.samples.iterators import (
    BorderedGroupedContentIterator
)
from report_writer.core.content.samples.iterators import ContentBlock
from report_writer.core.content.samples.iterators import ContentFilter
from report_writer.core.content.samples.iterators import ContentIterator
from report_writer.core.content.samples.iterators import ContentMapping
from report_writer.core.content.samples.iterators import ContentZip
from report_writer.core.content.samples.iterators import EnumerateContent
from report_writer.core.content.samples.iterators import GroupedContentIterator
from report_writer.core.content.samples.iterators import SizedContentIterator
from report_writer.core.content.samples.operators import BinaryOperatorContent
from report_writer.core.content.samples.operators import Content
from report_writer.core.content.samples.operators import UnaryOperatorContent
from report_writer.core.content.samples.statics import NoneContent
from report_writer.core.content.samples.statics import StaticContent
from report_writer.core.content.samples.wrappers import ConditionContent
from report_writer.core.content.samples.wrappers import ContentGate


class ContentTestCase(TestCase):
    def test_content_iterator(self):
        content_iterator = ContentIterator()
        context_list = [object() for _ in range(100)]

        result_list = list(content_iterator.render(context_list))
        self.assertEqual(context_list, result_list)

    def test_content_block(self):
        content_block = ContentBlock(
            [KeyContent('foo'), KeyContent('bar'), KeyContent('baz')])
        context = {'foo': 0, 'bar': 1, 'baz': 2}

        result_list = list(content_block.render(context))
        self.assertEqual(result_list, [0, 1, 2])

    def test_content_mapping(self):
        content_mapping = ContentMapping(
            [ContextContent(), StaticContent('bar')])
        context_list = ['foo', 'baz']

        result_list = list(content_mapping.render(context_list))
        self.assertEqual(result_list, ['foo', 'bar'])

    def test_none_content(self):
        content = NoneContent()
        context = object()

        result = content.render(context)
        self.assertIsNone(result)

    def test_static_content(self):
        content = StaticContent('foo')
        context = object()

        result = content.render(context)
        self.assertEqual(result, 'foo')

    def test_key_content(self):
        content = KeyContent('foo')
        context = {'foo': 0, 'bar': 1}

        result = content.render(context)
        self.assertEqual(result, 0)

        with self.assertRaises(KeyError):
            content.render({})

        content = KeyContent('foo', default=2)
        result = content.render({})
        self.assertEqual(result, 2)

    def test_content_gate(self):
        content = ContentGate(KeyContent('foo'), KeyContent('bar'))
        content0 = ContentGate(None, KeyContent('foo'))
        content1 = ContentGate(KeyContent('foo'), None)
        context = {'foo': {'bar': 2}}

        result = content.render(context)
        self.assertEqual(result, 2)

        result0 = content0.render(context)
        self.assertEqual(result0, {'bar': 2})

        result1 = content1.render(context)
        self.assertEqual(result1, {'bar': 2})

    def test_index_content(self):
        content = IndexContent(1)
        context = [0, 1]

        result = content.render(context)
        self.assertEqual(result, 1)

        with self.assertRaises(IndexError):
            content.render([])

        content = IndexContent(1, default=2)
        result = content.render([])
        self.assertEqual(result, 2)

    def test_attr_content(self):
        content = AttrContent('foo', 'bar')
        simple = type(str('A'), (), {})

        context = simple()
        setattr(context, 'foo', simple())
        setattr(getattr(context, 'foo'), 'bar', 1)

        result = content.render(context)
        self.assertEqual(result, 1)

        with self.assertRaises(AttributeError):
            content.render(simple())

        content = AttrContent('foo', 'bar', default=2)
        result = content.render(simple())
        self.assertEqual(result, 2)

    def test_template_content(self):
        content = TemplateContent('{{ foo }}{{ bar }}')
        context = {'foo': 0, 'bar': 1}

        result = content.render(context)
        self.assertEqual(result, '01')

    def test_sized_content(self):
        content = SizedContentIterator(size=3)
        context = range(100)

        result = tuple(content.render(context))
        self.assertEqual(result, (0, 1, 2))

    def test_bordered_iterator_content(self):
        content = BorderedContentIterator()
        context = range(5)

        result = tuple(content.render(context))
        self.assertEqual(result, (
            (0, True, False),
            (1, False, False),
            (2, False, False),
            (3, False, False),
            (4, False, True),
        ))

        context = range(1)

        result = tuple(content.render(context))
        self.assertEqual(result, (
            (0, True, True),
        ))

    def test_grouped_content(self):
        content = GroupedContentIterator(group_by=itemgetter(0))
        context = [
            (0, 0, 1),
            (0, 0, 2),
            (0, 1, 3),
            (1, 0, 4),
            (1, 1, 5),
            (1, 1, 6),
        ]

        result = content.render(context)
        self.assertEqual(
            tuple((k, list(g)) for k, g in result),
            (
                (0, [
                    (0, 0, 1),
                    (0, 0, 2),
                    (0, 1, 3),
                ]),
                (1, [
                    (1, 0, 4),
                    (1, 1, 5),
                    (1, 1, 6),
                ])
            )
        )

        content = GroupedContentIterator(
            GroupedContentIterator(group_by=itemgetter(1)),
            group_by=itemgetter(0),
            key_content=Content() >> str
        )

        result = content.render(context)
        self.assertEqual(
            tuple((k, [(k0, list(g0)) for k0, g0 in g]) for k, g in result),
            (
                ('0', [
                    (0, [
                        (0, 0, 1),
                        (0, 0, 2),
                    ]),
                    (1, [
                        (0, 1, 3),
                    ])
                ]),
                ('1', [
                    (0, [
                        (1, 0, 4),
                    ]),
                    (1, [
                        (1, 1, 5),
                        (1, 1, 6),
                    ]),
                ])
            )
        )

    def test_bordered_grouped_content(self):
        content = BorderedGroupedContentIterator(group_by=itemgetter(0))

        context = [
            (0, 0, 1),
            (0, 0, 2),
            (0, 1, 3),
            (1, 0, 4),
            (1, 1, 5),
            (1, 1, 6),
            (2, 0, 7),
        ]

        result = content.render(context)
        self.assertEqual(
            tuple(((k, list(g)), first, last)
                  for (k, g), first, last in result),
            (
                ((0, [
                    (0, 0, 1),
                    (0, 0, 2),
                    (0, 1, 3),
                ]), True, False),
                ((1, [
                    (1, 0, 4),
                    (1, 1, 5),
                    (1, 1, 6),
                ]), False, False),
                ((2, [
                    (2, 0, 7),
                ]), False, True)
            )
        )

    def test_content_zip(self):
        content = ContentZip(
            ContentBlock([IndexContent(0), IndexContent(1)]),
            [
                KeyContent('foo'),
                KeyContent('bar'),
            ]
        )

        context = {
            'foo': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'bar': [10, 11, 12, 13, 14, 15],
        }

        self.assertEqual(
            list(tuple(i) for i in content.render(context)),
            [
                (0, 10),
                (1, 11),
                (2, 12),
                (3, 13),
                (4, 14),
                (5, 15),
            ]
        )

        content = ContentZip(
            ContentBlock([IndexContent(0), IndexContent(1)]),
            [
                KeyContent('foo'),
                KeyContent('bar'),
            ],
            longest=True
        )

        self.assertEqual(
            list(tuple(i) for i in content.render(context)),
            [
                (0, 10),
                (1, 11),
                (2, 12),
                (3, 13),
                (4, 14),
                (5, 15),
                (6, None),
                (7, None),
                (8, None),
                (9, None),
            ]
        )

    def test_content_filter(self):
        base_content = Content() >> int
        content_filter = ContentFilter(
            base_content, filter_fn=lambda i: i % 2 == 0)
        content_filter_none = ContentFilter(base_content)
        context_list = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

        result_list = list(content_filter.render(context_list))
        self.assertEqual([0, 2, 4, 6, 8], result_list)

        result_list = list(content_filter_none.render(context_list))
        self.assertEqual([1, 2, 3, 4, 5, 6, 7, 8, 9], result_list)

    def test_condition_content(self):
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        content_true = StaticContent('true')
        content_false = StaticContent('false')

        condition = content0 > content1
        content = ConditionContent(condition, content_true, content_false)

        self.assertEqual(content.render((1, 2)), 'false')
        self.assertEqual(content.render((2, 1)), 'true')

    def test_enumerate_content(self):
        content0 = EnumerateContent()
        content1 = EnumerateContent(start=5)
        context = range(10)

        result0 = content0.render(context)
        result1 = content1.render(context)

        self.assertEqual(list(result0), [
            (0, 0),
            (1, 1),
            (2, 2),
            (3, 3),
            (4, 4),
            (5, 5),
            (6, 6),
            (7, 7),
            (8, 8),
            (9, 9),
        ])

        self.assertEqual(list(result1), [
            (5, 0),
            (6, 1),
            (7, 2),
            (8, 3),
            (9, 4),
            (10, 5),
            (11, 6),
            (12, 7),
            (13, 8),
            (14, 9),
        ])


class OperatorContentTest(TestCase):
    def test_unary_operator(self):
        content = UnaryOperatorContent(ContextContent(), str)
        result = content.render(1)

        self.assertIsInstance(result, str)
        self.assertEqual(result, '1')

    def test_binary_operator(self):
        content = BinaryOperatorContent(
            ContextContent(), StaticContent(2), lambda x, y: str(x + y))
        result = content.render(1)

        self.assertIsInstance(result, str)
        self.assertEqual(result, '3')

    def test_content_pos(self):
        content = Content()
        op_content = +content
        self.assertEqual(op_content.render(1), 1)

    def test_content_neg(self):
        content = Content()
        op_content = -content
        self.assertEqual(op_content.render(1), -1)

    def test_content_abs(self):
        content = Content()
        op_content = abs(content)
        self.assertEqual(op_content.render(-1), 1)

    def test_content_add(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 + content1
        self.assertEqual(op_content.render((1, 2)), 3)
        op_content = content + 2
        self.assertEqual(op_content.render(1), 3)

    def test_content_sub(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 - content1
        self.assertEqual(op_content.render((1, 2)), -1)
        op_content = content - 2
        self.assertEqual(op_content.render(1), -1)

    def test_content_mul(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 * content1
        self.assertEqual(op_content.render((1, 2)), 2)
        op_content = content * 2
        self.assertEqual(op_content.render(1), 2)

    def test_content_floordiv(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 // content1
        self.assertEqual(op_content.render((1, 2)), 0)
        op_content = content // 2
        self.assertEqual(op_content.render(1), 0)

    def test_content_truediv(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 / content1
        self.assertEqual(op_content.render((1, 2)), .5)
        op_content = content / 2
        self.assertEqual(op_content.render(1), .5)

    def test_content_mod(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 % content1
        self.assertEqual(op_content.render((1, 2)), 1)
        op_content = content % 2
        self.assertEqual(op_content.render(1), 1)

    def test_content_pow(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 ** content1
        self.assertEqual(op_content.render((1, 2)), 1)
        op_content = content ** 2
        self.assertEqual(op_content.render(1), 1)

    def test_content_lt(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 < content1
        self.assertTrue(op_content.render((1, 2)))
        self.assertFalse(op_content.render((3, 2)))
        op_content = content < 2
        self.assertTrue(op_content.render(1))
        self.assertFalse(op_content.render(3))

    def test_content_le(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 <= content1
        self.assertTrue(op_content.render((1, 2)))
        self.assertTrue(op_content.render((2, 2)))
        self.assertFalse(op_content.render((3, 2)))
        op_content = content <= 2
        self.assertTrue(op_content.render(1))
        self.assertTrue(op_content.render(2))
        self.assertFalse(op_content.render(3))

    def test_content_gt(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 > content1
        self.assertFalse(op_content.render((1, 2)))
        self.assertTrue(op_content.render((3, 2)))
        op_content = content > 2
        self.assertFalse(op_content.render(1))
        self.assertTrue(op_content.render(3))

    def test_content_ge(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 >= content1
        self.assertFalse(op_content.render((1, 2)))
        self.assertTrue(op_content.render((2, 2)))
        self.assertTrue(op_content.render((3, 2)))
        op_content = content >= 2
        self.assertFalse(op_content.render(1))
        self.assertTrue(op_content.render(2))
        self.assertTrue(op_content.render(3))

    def test_content_eq(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 == content1
        self.assertFalse(op_content.render((1, 2)))
        self.assertTrue(op_content.render((2, 2)))
        op_content = content == 2
        self.assertFalse(op_content.render(1))
        self.assertTrue(op_content.render(2))

    def test_content_ne(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 != content1
        self.assertTrue(op_content.render((1, 2)))
        self.assertFalse(op_content.render((2, 2)))
        op_content = content != 2
        self.assertTrue(op_content.render(1))
        self.assertFalse(op_content.render(2))

    def test_content_or(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 | content1
        self.assertTrue(op_content.render((True, False)))
        self.assertFalse(op_content.render((False, False)))
        op_content = content | False
        self.assertTrue(op_content.render(True))
        self.assertFalse(op_content.render(False))

    def test_content_and(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 & content1
        self.assertTrue(op_content.render((True, True)))
        self.assertFalse(op_content.render((True, False)))
        op_content = content & True
        self.assertTrue(op_content.render(True))
        self.assertFalse(op_content.render(False))

    def test_content_xor(self):
        content = Content()
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        op_content = content0 ^ content1
        self.assertTrue(op_content.render((True, False)))
        self.assertFalse(op_content.render((True, True)))
        op_content = content ^ True
        self.assertTrue(op_content.render(False))
        self.assertFalse(op_content.render(True))

    def test_content_not(self):
        content = Content()
        op_content = ~content
        self.assertEqual(op_content.render(True), False)
        self.assertEqual(op_content.render(False), True)

    def test_content_invert(self):
        content = Content()
        op_content = content.invert()
        self.assertEqual(op_content.render(1), ~1)
        self.assertEqual(op_content.render(2), ~2)
        self.assertEqual(op_content.render(3), ~3)

    def test_pass_content_as_argument(self):
        content = Content()
        op_content = content >> str
        result = op_content.render(1)
        self.assertIsInstance(result, str)
        self.assertEqual(result, '1')

    def test_content_shift(self):
        content = Content()
        lshift_content = content << 2
        rshift_content = content << -2

        self.assertEqual(lshift_content.render(1), 4)
        self.assertEqual(rshift_content.render(4), 1)

        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))

        shift_content = content0 << content1

        self.assertEqual(shift_content.render((1, 2)), 4)
        self.assertEqual(shift_content.render((4, -2)), 1)

    def test_complex_operations(self):
        content0 = Content(IndexContent(0))
        content1 = Content(IndexContent(1))
        content2 = Content(IndexContent(2))
        math_content = -((content0 + content1) - content2)
        self.assertEqual(math_content.render((2, 3, 4)), -1)
        bool_content = ~((content0 & content1) | content2)
        self.assertFalse(bool_content.render((True, False, True)))
