from itertools import islice
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Tuple

from report_writer.core.report_objects.base import BaseReportObject
from report_writer.core.report_objects.cursor import Cursor


if TYPE_CHECKING:
    from report_writer.core.report_objects.cursor import CursorFrame
    HeaderFactory = Callable[[int, int], BaseReportObject]


class ObjectIterator(BaseReportObject):
    """
    Последовательность объектов

    Для каждого элемента последовательности данных записывает
    базовый объект отчёта

    Может быть использовано для строк произвольной длины (тогда в качестве
    report_object используется клетка отчёта) или таблиц
    (тогда - строка отчёта).

    :param report_object: Объект-элемент последовательности
    :param direction: Направление записи объектов
    """
    def __init__(
            self,
            report_object: BaseReportObject,
            direction: Tuple[int, int]
    ) -> None:
        self._report_object: BaseReportObject = report_object
        self._direction: Tuple[int, int] = direction

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL
    ) -> None:
        with cursor.get_state() as state:
            for _data in data:
                self._report_object.write_object(
                    cursor, _data, self._direction)

            state.frame.close(direction)


class BorderedObjectIterator(ObjectIterator):
    """
    Последовательность объектов у которого можно задать отдельные
    объекты для граничных значений.

    Полезно если у граничных элементов должно быть другое форматирование.

    :param report_object: Объект-элемент последовательности
    :param direction: Направление записи объектов
    :param first_object: Первый объект-элемент последовательности
    :param last_object: Последний объект-элемент последовательности
    :param single_object: Единственный объект-элемент последовательности
    """
    def __init__(
            self,
            report_object: BaseReportObject,
            direction: Tuple[int, int],
            first_object: Optional[BaseReportObject] = None,
            last_object: Optional[BaseReportObject] = None,
            single_object: Optional[BaseReportObject] = None
    ) -> None:
        super(BorderedObjectIterator, self).__init__(report_object, direction)
        self._first_object: BaseReportObject = first_object or report_object
        self._last_object: BaseReportObject = last_object or report_object
        self._single_object: BaseReportObject = (
                single_object or self._first_object)

    def _get_report_object(self, first: bool, last: bool) -> BaseReportObject:
        """
        Получение объекта для элемента итератора

        :param first: Признак первого объекта
        :param last: Признак последнего объекта
        """
        if first and last:
            obj = self._single_object
        elif first:
            obj = self._first_object
        elif last:
            obj = self._last_object
        else:
            obj = self._report_object

        return obj

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL
    ) -> None:
        with cursor.get_state() as state:
            for _data, first, last in data:
                self._get_report_object(first, last).write_object(
                    cursor, _data, self._direction)

            state.frame.close(direction)


class ObjectBlock(BaseReportObject):
    """
    Блок объектов.

    Для каждого элемента последовательности данных записывает
    соответствующий (по порядку) объект отчёта.

    Может быть использовано для смешанных строк определённой длины (тогда
    report_objects - итератор клеток отчёта для каждого элемента строки).

    :param report_objects: Последовательность объектов блока
    :param direction: Направление записи объектов
    """
    def __init__(
            self,
            report_objects: Iterable[BaseReportObject],
            direction: Tuple[int, int]
    ) -> None:
        self._report_objects: Iterable[BaseReportObject] = report_objects
        self._direction: Tuple[int, int] = direction

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL,
    ) -> None:
        with cursor.get_state() as state:
            for obj, _data in zip(self._report_objects, data):
                obj.write_object(cursor, _data, self._direction)

            state.frame.close(direction)


class SizedObjectIterator(ObjectIterator):
    """
    Последовательность объектов фиксированного размера.

    Может быть использовано для однородных строк определённой длины (тогда
    report_object - клетка отчёта используемая во всех элментах строки).

    :param report_object: Объект-элемент последовательности
    :param direction: Направление записи объектов
    :param size: Размер последовательности
    """
    def __init__(
            self,
            report_object: BaseReportObject,
            direction: Tuple[int, int],
            size: Optional[int] = None
    ) -> None:
        super(SizedObjectIterator, self).__init__(report_object, direction)
        self._size: Optional[int] = size

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL,
    ) -> None:
        data = islice(data, self._size)
        super(SizedObjectIterator, self).write_object(cursor, data, direction)


class ObjectIteratorWithHeaders(ObjectIterator):
    """
    Последовательность объектов отчёта с заголовками.

    Записывает объекты в указанном направлении, добавляя потом заголовки
    для них.

    Может быть использовано для сгруппированных таблиц
    (в этом случае report_object - объект для группы,
    header_object - заголовок группы), причём объектом для группы может так
    же быть сгруппированная таблица, что позволяет создавать сгруппированные
    таблицы любой вложенности.

    :param report_object: Объект группы
    :param header_object: Фабрика для объектов заголовков групп
                          (на вход подаётся высота и ширина фрейма группы)
    :param direction: Направление записи
    :param header_size: Общий размер заголовка
    """
    def __init__(self,
                 report_object: BaseReportObject,
                 header_object: 'HeaderFactory',
                 direction: Tuple[int, int],
                 header_size: int = 1
                 ) -> None:
        super(ObjectIteratorWithHeaders, self).__init__(
            report_object, direction)
        self._header_object: 'HeaderFactory' = header_object
        self._header_size: int = header_size

    @staticmethod
    def make_header_factory(
        multiline_header_factory: 'HeaderFactory',
        single_line_header_factory: Callable[[], BaseReportObject]
    ) -> 'HeaderFactory':
        """
        Метод для создания дефолтной фабрики объектов заголовков групп.

        Подходит для большинства случаев.

        Фабрика создаётся на основе фабрик объединённой и простой клетки.
        Для объединённой передаётся размер клетки (кортеж (высота, ширина)),
        для простой дополнительных параметров не передаётся

        :param multiline_header_factory: Фабрика для многострочного заголовка
        :param single_line_header_factory: Фабрика для однострочного заголовка
        :return: Фабрика для заголовка
        """
        def factory(frame_h: int, frame_w: int) -> BaseReportObject:
            if max(frame_h, frame_w) > 1:
                return multiline_header_factory(
                    max(frame_h, 1), max(frame_w, 1))
            else:
                return single_line_header_factory()

        return factory

    def _write_header(
            self,
            cursor: Cursor,
            group_frame: 'CursorFrame',
            data: Any
    ) -> None:
        """
        Запись заголовка группы.

        :param cursor: Используемый курсор
        :param group_frame: Фрейм записи объекта-группы
        :param data: Данные для заголовка
        """
        if not group_frame.has_points:
            return

        key, _ = data
        frame_h, frame_w = self.calculate_group_header_dimensions(group_frame)
        self._header_object(frame_h, frame_w).write_object(
            cursor, key, Cursor.invert_direction(self._direction))

    def _write_group(self, cursor: Cursor, data: Any) -> None:
        """
        Запись группы.

        :param cursor: Используемый курсор
        :param data: Данные для группы
        """
        _, group = data
        self._report_object.write_object(cursor, group, self._direction)

    def calculate_group_header_dimensions(
            self,
            group_frame: 'CursorFrame'
    ) -> Tuple[int, int]:
        """
        Высчитывание размера заголовка для фрейма группы.

        :param group_frame: Фрейм записи объекта-группы
        :return: Размеры заголовка
        """
        frame_h, frame_w = group_frame.height, group_frame.width

        rows, cols = self._direction

        frame_h *= rows
        frame_w *= cols

        return frame_h, frame_w

    def _write_item(self, cursor: Cursor, data: Any) -> None:
        """
        Запись элемента итератора.

        :param cursor: Используемый курсор
        :param data: Данные элемента
        """
        with cursor.get_state() as group_outer_state:
            cursor.move(
                Cursor.invert_direction(self._direction), self._header_size)

            with cursor.get_state() as group_inner_state:
                self._write_group(cursor, data)

            cursor.back()

            self._write_header(cursor, group_inner_state.frame, data)

            group_outer_state.frame.close(self._direction)

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL,
    ) -> None:
        with cursor.get_state() as state:
            for _data in data:
                self._write_item(cursor, _data)

            state.frame.close(direction)


class BorderedObjectIteratorWithHeaders(ObjectIteratorWithHeaders):
    """
    Последовательность объектов с заголовками с возможностью задавать
    отдельные объекты для граничных групп и их заголовков.

    Полезно если у граничных элементов должно быть другое форматирование.

    :param report_object: Объект группы
    :param header_object: Фабрика для объектов заголовков групп
                          (на вход подаётся высота и ширина фрейма группы)
    :param direction: Направление записи
    :param header_size: Общий размер заголовка
    :param first_object: Объект группы (первая группа)
    :param last_object: Объект группы (последняя группа)
    :param single_object: Объект группы (единственная группа)
    :param first_header_object: Фабрика для объектов заголовков групп
                                (первая группа)
    :param last_header_object: Фабрика для объектов заголовков групп
                               (последняя группа)
    :param single_header_object: Фабрика для объектов заголовков групп
                                 (единственная группа)
    """
    def __init__(
        self,
        report_object: BaseReportObject,
        header_object: 'HeaderFactory',
        direction: Tuple[int, int],
        header_size: int = 1,
        first_object: Optional[BaseReportObject] = None,
        last_object: Optional[BaseReportObject] = None,
        single_object: Optional[BaseReportObject] = None,
        first_header_object: Optional['HeaderFactory'] = None,
        last_header_object: Optional['HeaderFactory'] = None,
        single_header_object: Optional['HeaderFactory'] = None
    ) -> None:
        super(BorderedObjectIteratorWithHeaders, self).__init__(
            report_object, header_object, direction, header_size)

        self._first_object: BaseReportObject = (
            first_object or self._report_object)
        self._last_object: BaseReportObject = (
            last_object or self._report_object)
        self._single_object: BaseReportObject = (
            single_object or self._first_object)
        self._first_header_object: 'HeaderFactory' = (
            first_header_object or self._header_object)
        self._last_header_object: 'HeaderFactory' = (
            last_header_object or self._header_object)
        self._single_header_object: 'HeaderFactory' = (
                single_header_object or self._first_header_object)

    def _write_header(
            self,
            cursor: Cursor,
            group_frame: 'CursorFrame',
            data: Any
    ) -> None:
        if not group_frame.has_points:
            return None

        (key, _), first, last = data

        if first and last:
            obj = self._single_header_object
        elif first:
            obj = self._first_header_object
        elif last:
            obj = self._last_header_object
        else:
            obj = self._header_object

        frame_h, frame_w = self.calculate_group_header_dimensions(group_frame)
        obj(frame_h, frame_w).write_object(
            cursor, key, Cursor.invert_direction(self._direction))

    def _write_group(self, cursor: Cursor, data: Any) -> None:
        (_, group), first, last = data

        if first and last:
            obj = self._single_object
        elif first:
            obj = self._first_object
        elif last:
            obj = self._last_object
        else:
            obj = self._report_object

        obj.write_object(cursor, group, self._direction)
