import datetime
import os
from tempfile import NamedTemporaryFile
from unittest import TestCase

from openpyxl import Workbook
from openpyxl import load_workbook
from openpyxl.styles import Font
from openpyxl.styles import NamedStyle

from report_writer.core.report_objects.cursor import Cursor
from report_writer.core.report_objects.samples.simple import Cell
from report_writer.impl.xlsx.report.report_file import ReportFile
from report_writer.impl.xlsx.report.writers.format import ColumnStyleWriter
from report_writer.impl.xlsx.report.writers.format import ColumnWidthWriter
from report_writer.impl.xlsx.report.writers.format import NamedStyleWriter
from report_writer.impl.xlsx.report.writers.format import RowHeightWriter
from report_writer.impl.xlsx.report.writers.format import RowStyleWriter
from report_writer.impl.xlsx.report.writers.format import StyledCellWriter
from report_writer.impl.xlsx.report.writers.format import (
    StyledMergedCellWriter
)
from report_writer.impl.xlsx.report.writers.merge import MergedCellWriter
from report_writer.impl.xlsx.report.writers.merge import MergeWriter
from report_writer.impl.xlsx.report.writers.simple import CellWriter
from report_writer.utils.read_utils import read_xlsx


# На Windows файл нельзя будет удалить
_DELETE_TMP_FILE = os.name != 'nt'


class OpenpyxlReportFileTest(TestCase):
    def test_report_file_saving(self):
        with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                delete=_DELETE_TMP_FILE) as tmp:
            report = ReportFile(tmp.name)
            cell = Cell(CellWriter(report))
            cursor = Cursor(0, 0)

            cell.write_object(cursor, 1, Cursor.VERTICAL)

            report.save()

            wb = load_workbook(tmp.name)

            self.assertEqual(next(read_xlsx(wb)), [[1]])

    def test_report_templates(self):
        with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                delete=_DELETE_TMP_FILE) as template:
            wb = Workbook()
            wb.active['A1'] = '{{ string }}'
            wb.active['A2'] = '{{ foo.bar }}'
            wb.active['A3'] = '{{ boolean }}'
            wb.active['A4'] = '{{ number }}'
            wb.active['A5'] = '{{ date }}'
            wb.active['A6'] = 1
            wb.save(template.name)

            with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                    delete=_DELETE_TMP_FILE) as tmp:
                report = ReportFile(tmp.name, template.name, {
                    'string': 'string',
                    'foo': {'bar': 0},
                    'boolean': True,
                    'number': 1,
                    'date': datetime.date(1970, 1, 1)
                })

                report.save()

                wb = load_workbook(tmp.name)
                self.assertEqual(next(read_xlsx(wb)), [
                    ['string'],
                    ['0'],
                    ['True'],
                    ['1'],
                    ['1970-01-01'],
                    [1],
                ])

    def test_report_cursors(self):
        with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                delete=_DELETE_TMP_FILE) as template:
            wb = Workbook()

            wb.active.title = 'Foo'

            wb.active['A1'].value = 'foo'
            wb.active['A2'].value = 'cursor'
            wb.active['A3'].value = 'cursor#'
            wb.active['B2'].value = 'cursor#[main]'

            wb.save(template.name)

            report = ReportFile(None, template.name, {})

            self.assertEqual(len(report._template_file._all_cursors['Foo']), 1)
            self.assertIn('main', report._template_file._all_cursors['Foo'])

            cursor = report.get_cursor('main')

            self.assertEqual(cursor.point, (1, 1))

            report = ReportFile(None)
            with self.assertRaises(ValueError):
                report.get_cursor('Foo')

    def test_multiple_sheet_cursors(self):
        with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                delete=_DELETE_TMP_FILE) as template:
            wb = Workbook()

            wb.active.title = 'Foo'

            wb.active['A1'].value = 'foo'
            wb.active['A2'].value = 'cursor'
            wb.active['A3'].value = 'cursor#'
            wb.active['B2'].value = 'cursor#[main]'

            wb.active = wb.create_sheet('Bar')

            wb.active['A1'].value = 'foo'
            wb.active['A2'].value = 'cursor'
            wb.active['A3'].value = 'cursor#'
            wb.active['C3'].value = 'cursor#[foo]'

            wb.save(template.name)

            report = ReportFile(None, template.name, {})

            self.assertEqual(len(report._template_file._all_cursors['Foo']), 1)
            self.assertIn('main', report._template_file._all_cursors['Foo'])
            self.assertNotIn('foo', report._template_file._all_cursors['Foo'])

            self.assertEqual(len(report._template_file._all_cursors['Bar']), 1)
            self.assertIn('foo', report._template_file._all_cursors['Bar'])
            self.assertNotIn('main', report._template_file._all_cursors['Bar'])

            report.set_worksheet('Foo')
            cursor = report.get_cursor('main')
            self.assertEqual(cursor.point, (1, 1))

            report.set_worksheet('Bar')
            cursor = report.get_cursor('foo')
            self.assertEqual(cursor.point, (2, 2))

    def test_report_file_add_named_style(self):
        report = ReportFile(None)
        style = NamedStyle('foo')
        report.add_named_style(style)

        self.assertIn(style, report.workbook._named_styles)

    def test_report_set_worksheet(self):
        with NamedTemporaryFile(suffix='.xlsx', mode='w',
                                delete=_DELETE_TMP_FILE) as tmp:
            report = ReportFile(tmp.name)

            report.set_worksheet('b')
            report.set_worksheet('c')
            report.set_worksheet('a', idx=0)
            report.set_worksheet('b')

            cell = Cell(CellWriter(report))
            cursor = Cursor(0, 0)
            cell.write_object(cursor, 1, Cursor.VERTICAL)

            report.save()

            wb = load_workbook(tmp.name)
            self.assertEqual(list(read_xlsx(wb)), [
                [],
                [],
                [[1]],
                [],
            ])


class OpenpyxlReportObjectsTest(TestCase):
    def test_cell_object(self):
        report = ReportFile(None)
        cursor = Cursor(0, 0)
        cell = Cell(CellWriter(report))

        cell.write_object(cursor, 1, Cursor.VERTICAL)

        self.assertEqual(next(read_xlsx(report.workbook)), [[1]])

        cell.write_object(cursor, 2, Cursor.VERTICAL)
        cell.write_object(cursor, 3, Cursor.HORIZONTAL)
        cell.write_object(cursor, 4, Cursor.HORIZONTAL)
        cell.write_object(cursor, 5, Cursor.HORIZONTAL)

        self.assertEqual(
            next(read_xlsx(report.workbook)),
            [
                [1, None, None],
                [2, None, None],
                [3, 4,    5]
            ]
        )

    def test_merge_object(self):
        report = ReportFile(None)
        cursor = Cursor(0, 0)
        cell = Cell(CellWriter(report))
        merge = Cell(MergeWriter(report, 2, 3), 0, 0)

        merge.write_object(cursor, None, Cursor.VERTICAL)
        cell.write_object(cursor, 1, Cursor.VERTICAL)

        self.assertEqual(
            next(read_xlsx(report.workbook)),
            [
                [1,    None, None],
                [None, None, None],
            ]
        )

    def test_merged_cell_object(self):
        report = ReportFile(None)
        cursor = Cursor(0, 0)
        cell = Cell(MergedCellWriter(report, 2, 3), 2, 3)

        cell.write_object(cursor, 1, Cursor.VERTICAL)
        cell.write_object(cursor, 2, Cursor.HORIZONTAL)
        cell.write_object(cursor, 3, Cursor.HORIZONTAL)

        self.assertEqual(
            next(read_xlsx(report.workbook)),
            [
                [1,    None, None, None, None, None],
                [None, None, None, None, None, None],
                [2,    None, None, 3,    None, None],
                [None, None, None, None, None, None],
            ]
        )

    def test_named_style_writer(self):
        report = ReportFile(None)

        style0 = NamedStyle('foo')
        style1 = NamedStyle('bar')

        NamedStyleWriter(report, style0).write()
        NamedStyleWriter(report, style1).write()

        self.assertIn(style0, report.workbook._named_styles)
        self.assertIn(style1, report.workbook._named_styles)

    def test_formatted_cell(self):
        report = ReportFile(None)
        cursor = Cursor(0, 0)
        cell = Cell(StyledCellWriter(report, style='foo'))
        report.add_named_style(NamedStyle('foo'))

        cell.write_object(cursor, 0, Cursor.VERTICAL)

        self.assertEqual(report.workbook.active['A1'].style, 'foo')

    def test_formatted_merged_cell(self):
        report = ReportFile(None)
        cursor = Cursor(0, 0)
        cell = Cell(StyledMergedCellWriter(report, 2, 3, style='foo'), 2, 3)
        report.add_named_style(NamedStyle('foo'))

        cell.write_object(cursor, 0, Cursor.VERTICAL)

        self.assertEqual(report.workbook.active['A1'].style, 'foo')
        self.assertEqual(
            next(read_xlsx(report.workbook)),
            [
                [0,    None, None],
                [None, None, None],
            ]
        )

    def test_column_width_writer(self):
        report = ReportFile(None)
        ColumnWidthWriter(report).write(0, 0, 10)
        ColumnWidthWriter(report, 4).write(0, 2, 15)

        wb = report.workbook

        self.assertAlmostEqual(
            wb.active.column_dimensions['A'].width, 10, delta=1)
        self.assertEqual(wb.active.column_dimensions['C'].max, 6)
        self.assertAlmostEqual(
            wb.active.column_dimensions['C'].width, 15, delta=1)

    def test_column_style_writer(self):
        report = ReportFile(None)
        writer0 = ColumnStyleWriter(report, font=Font(bold=True))
        writer1 = ColumnStyleWriter(report, 10, font=Font(bold=True))

        writer0.write(0, 0)
        writer1.write(0, 5)

        wb = report.workbook

        self.assertTrue(wb.active.column_dimensions['A'].font.b)
        self.assertTrue(wb.active.column_dimensions['F'].font.b)
        self.assertEqual(wb.active.column_dimensions['F'].max, 15)

    def test_row_height_writer(self):
        report = ReportFile(None)
        RowHeightWriter(report).write(0, 0, 100)

        wb = report.workbook
        self.assertEqual(wb.active.row_dimensions[1].r, 1)
        self.assertEqual(wb.active.row_dimensions[1].height, 100)

    def test_row_style_writer(self):
        report = ReportFile(None)
        writer0 = RowStyleWriter(report, font=Font(bold=True))

        writer0.write(0, 0)

        wb = report.workbook

        self.assertTrue(wb.active.row_dimensions[1].font.b)
