from unittest import TestCase

from report_writer.core.report_objects.cursor import Cursor
from report_writer.core.report_objects.samples.simple import Cell
from report_writer.utils.iterators import with_next
from report_writer.utils.mock_report import CellWriter
from report_writer.utils.mock_report import FormattedCellWriter
from report_writer.utils.mock_report import FormattedMergedCellWriter
from report_writer.utils.mock_report import MergedCellWriter
from report_writer.utils.mock_report import NamedFormatWriter
from report_writer.utils.mock_report import ReportFileMock


class IteratorsTest(TestCase):
    def test_with_next_iterator(self):
        iterable = [0, 1, 2, 3, 4, 5, 6]
        result = list(with_next(iterable))

        self.assertEqual(
            result,
            [
                (0, 1),
                (1, 2),
                (2, 3),
                (3, 4),
                (4, 5),
                (5, 6),
                (6, None),
            ]
        )

        iterable = []
        result = list(with_next(iterable))

        self.assertEqual(result, [])


class MockReportTest(TestCase):
    def test_cell_object(self):
        report = ReportFileMock(2, 2)
        cell = Cell(CellWriter(report))

        with Cursor(0, 0).get_state() as state:
            cell.write_object(state.cursor, 1, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1, None],
            [None, None]
        ])

    def test_cell_cursor(self):
        report = ReportFileMock(3, 3)
        cell = Cell(CellWriter(report))

        with Cursor(1, 1).get_state() as state:
            cell.write_object(state.cursor, 1, Cursor.HORIZONTAL)
            cell.write_object(state.cursor, 2, Cursor.VERTICAL)
            cell.write_object(state.cursor, 3, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [None, None, None],
            [None, 1,    2],
            [None, None, 3],
        ])

    def test_merged_cell_object(self):
        report = ReportFileMock(3, 3)
        cell = Cell(MergedCellWriter(report, 2, 2), 2, 2)

        with Cursor(0, 0).get_state() as state:
            cell.write_object(state.cursor, 1, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1, None, None],
            [None, None, None],
            [None, None, None],
        ])

        self.assertIn({(0, 0), (0, 1), (1, 0), (1, 1)}, report.merged)

    def test_merged_cell_cursor(self):
        report = ReportFileMock(5, 5)
        cell = Cell(MergedCellWriter(report, 2, 2), 2, 2)

        with Cursor(0, 0).get_state() as state:
            cell.write_object(state.cursor, 1, Cursor.HORIZONTAL)
            cell.write_object(state.cursor, 2, Cursor.VERTICAL)
            cell.write_object(state.cursor, 3, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [1, None, 2, None, None],
            [None, None, None, None, None],
            [None, None, 3, None, None],
            [None, None, None, None, None],
            [None, None, None, None, None],
        ])

        self.assertIn({(0, 0), (0, 1), (1, 0), (1, 1)}, report.merged)
        self.assertIn({(0, 2), (0, 3), (1, 2), (1, 3)}, report.merged)
        self.assertIn({(2, 2), (2, 3), (3, 2), (3, 3)}, report.merged)

    def test_formatted_cell(self):
        report = ReportFileMock(2, 2)
        cell = Cell(FormattedCellWriter(report, 'format'))

        with Cursor(0, 0).get_state() as state:
            NamedFormatWriter(report, 'format', 0).write()
            cell.write_object(state.cursor, 1, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [(0, 1), None],
            [None, None]
        ])

    def test_formatted_merged_cell(self):
        report = ReportFileMock(3, 3)
        cell = Cell(FormattedMergedCellWriter(report, 2, 2, 'format'), 2, 2)

        with Cursor(0, 0).get_state() as state:
            NamedFormatWriter(report, 'format', 0).write()
            cell.write_object(state.cursor, 1, Cursor.VERTICAL)

        self.assertEqual(report.data, [
            [(0, 1), None, None],
            [None, None, None],
            [None, None, None],
        ])

        self.assertIn({(0, 0), (0, 1), (1, 0), (1, 1)}, report.merged)
